#include <cmath>
#ifndef TEST


/*
 * BulbState
 *
 *   The state of the lightbulb
 */

enum BulbState { OFF = 0, ON = 1 };


/*
 * ExternalState
 *
 *   Environmental data which the prisoners are aware of and/or are able
 * to influence: the state of the lightbulb and the count of days.
 */

class ExternalState {
public:
	ExternalState() { _day = 0; _bulb = OFF; }
	
	inline int getDay() const { return _day; }
	inline BulbState getLightbulb() const { return _bulb; }
	
	inline void setLightbulb(BulbState state) { _bulb = state; }

protected:
	friend class Simulator;
	inline void nextDay() { ++_day; }

private:
	int _day;
	BulbState _bulb;
};


/*
 * SharedKnowledge
 *
 *   A representation of the strategy meeting: number of days in each
 * round, number token required to fill each badge, etc.
 */

class SharedKnowledge {
public:
	inline static int getNumBadges() { return _numBadges; }
	inline static int getBadgeSize(int badge) { return _badgeSizes[badge]; }
	inline static int getNumRounds() { return _numRounds; }
	inline static int getNumRepeatedRounds() { return _numRepeatedRounds; }
	inline static int getRoundLength(int round) { return _roundLengths[round]; }
	
	// computational shortcuts to avoid the selected prisoner having to
	// recompute the current round from the day number every day
	inline static int getCurrentRound(int day) {
		if (day < _currentStart) {
			_currentRound = -1; _currentStart = 0; _currentEnd = 1;
		}
		while (day >= _currentEnd) {
			if (++_currentRound == _numRounds) {
				_currentRound -= _numRepeatedRounds;
			}
			_currentStart = _currentEnd;
			_currentEnd += _roundLengths[_currentRound];
		}
		return _currentRound;
	}
	inline static int getCurrentRoundDay(int day) {
		return day - _currentStart + 1;
	}
	
private:
	static int _numBadges;
	static int _badgeSizes[];
	static int _numRounds;
	static int _numRepeatedRounds;
	static int _roundLengths[];
	
	static int _currentRound;
	static int _currentStart;
	static int _currentEnd;
};


/*
 * InternalState
 *
 *   Each prisoner's mental list: numbers of tokens, badges, etc.
 */

class InternalState {
public:
	InternalState() {
		_badgeSizes = new int[SharedKnowledge::getNumBadges()];
		reset();
	}
	~InternalState() { delete[] _badgeSizes; }
	
	inline bool hasTokens() const { return _numTokens > 0; }
	inline bool hasBadges() const { return _numBadges > 0; }
	inline bool hasFilledBadges() const { return _filledBadges > 0; }
	inline bool hasCrown() const { return _crown; }
	inline bool victory() const {
		return _filledBadges == SharedKnowledge::getNumBadges();
	}
	
	inline void assignBadge(int badge) {
		_badgeSizes[_numBadges] = SharedKnowledge::getBadgeSize(badge);
		++_numBadges;
	}
	inline void assignCrown() { _crown = true; }
	inline void pickupTokens(int tokens) {
		_numTokens += tokens;
		while (_numBadges > 0 && _numTokens >= _badgeSizes[_numBadges - 1]) {
			_numTokens -= _badgeSizes[_numBadges - 1];
			--_numBadges; ++_filledBadges;
		}
	}
	inline void pickupBadge() { ++_filledBadges; }
	inline void dropToken() { --_numTokens; }
	inline void dropBadge() { --_filledBadges; }

	void reset() {
		_numTokens = 1;
		_numBadges = _filledBadges = 0;
		_crown = false;
	}

private:
	int _numTokens, _numBadges, _filledBadges;
	int * _badgeSizes;
	bool _crown;
};


/*
 * Prisoner
 *
 *   A prisoner together with their internal state.  The declaresVictory
 * function is the main decision point for all prisoner actions.
 */

class Prisoner {
public:
	Prisoner();
	
	bool declaresVictory(ExternalState & extState);
	
	void reset();
	
private:
	InternalState _intState;
};


#endif


/*
 * MersenneTwister
 *
 *   A utility class implementing a pseudorandom number generator of
 * sufficient quality for Monte Carlo simulations (unlike the rand()
 * library function which may show considerable artifacting).
 *
 *   Returns an integer uniformly distributed in 0 to max-1
 */

class MersenneTwister {
public:
	MersenneTwister(unsigned int max, unsigned int seed = 0);
	
	unsigned int rand();
	
	void reseed(unsigned int seed = 0);

private:
	void _generate();
	
	const unsigned int _max, _maxrand;
	unsigned int _state[624];
	unsigned int _index;
};


/*
 * Statistics
 *
 *   A utility class implementing a simple accumulator for statistical
 * data about a set of simulations
 */

class Statistics {
public:
	Statistics() { _m = _s = _n = 0; }
	
	inline int getSamples() const { return _n; }
	inline double getMean() const { return _m; }
	inline double getStdDev() const {
		return (_n > 1) ? std::sqrt(_s / (_n - 1)) : 0.0;
	}
	
	inline void addSample(int value) {
		double x = value, oldM = _m;
		_m += (x - oldM)/(++_n);
		_s += (x - oldM)*(x - _m);
	}
	
private:
	int _n;
	double _m, _s;
};


#ifndef TEST


/*
 * Simulator
 *
 *   The scenario simulator: simulates a single run or multiple runs of
 * the prisoners' decision algorithm to completion and counts the number
 * of days required
 */

class Simulator {
public:
	Simulator(int prisoners, int prng_seed = 0);
	~Simulator();

	int simulate();
	Statistics simulate(int runs);
	
	void reseed(int prng_seed = 0);

private:
	const int _numPrisoners;
	Prisoner * _prisoners;
	bool * _visited;
	MersenneTwister _prng;
};


#endif
